using System;
using System.Collections;
using System.Collections.Specialized;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Xml;
using System.Xml.Serialization;

using Microsoft.Win32;

using GBPVR.Public;

namespace RecorderHelper
{
	/// <summary>
	/// Summary description for Config.
	/// </summary>
	[XmlType("settings")]
	public class Config
	{

		#region Private Variables
		
		[NonSerialized()]
		private HybridDictionary globalSettingNodes;
		[NonSerialized()]
		private HybridDictionary captureSourceNodes;
		
		[NonSerialized()]
		private string configFileName;
		
		[NonSerialized()]
		private XmlDocument doc;
		
		#endregion

		#region Public Global Settings Methods

		public string GetSetting(object moduleObject, string settingName)
		{
			string moduleName = moduleObject.GetType().Assembly.GetName().Name;

			Logger.Verbose("RecorderHelperConfig: GetSettings(" + moduleName + ", " + settingName + ")");

			LoadFromFile(false);

			XmlNode globalSettingsNode = (XmlNode)globalSettingNodes[moduleName];

			if (globalSettingsNode == null) 
			{
				Logger.Verbose("RecorderHelperConfig: GetSettings, globalSettingsNode == null");
				return null;
			}

			XmlNode settingNode = globalSettingsNode.SelectSingleNode(settingName);

			if (settingNode == null) 
			{
				Logger.Verbose("RecorderHelperConfig: GetSettings, settingNode == null");
				return null;
			}

			if (settingNode.InnerText != null)
				Logger.Verbose("RecorderHelperConfig: GetSettings, return '" + settingNode.InnerText + "'");
			else
				Logger.Verbose("RecorderHelperConfig: GetSettings, return null");

			return settingNode.InnerText;
		}

		#endregion

		#region Public Capture Source Methods
        
		public void Load(int captureSourceOID, IRecordingSourceConfiguration config)
		{
			Logger.Verbose("RecorderHelperConfig: Load(" + captureSourceOID.ToString() + ")");

			LoadFromFile(false);

			XmlNode captureSourceNode = (XmlNode)captureSourceNodes[captureSourceOID];

			if (captureSourceNode != null)
			{
				CopyNodeToObject(captureSourceNode, config);
			}
			else
			{
				Logger.Verbose("RecorderHelperConfig: Load(" + captureSourceOID.ToString() + "): Missing!");
			}
		}

		public void Save(int captureSourceOID, IRecordingSourceConfiguration config)
		{
			Logger.Verbose("RecorderHelperConfig: Save(" + captureSourceOID.ToString() + ")");

			Mutex mtx = new Mutex(false, "RecorderHelper");
			if (!mtx.WaitOne(5000, false))
				throw new ApplicationException("Could not acquire RecorderHelper mutex");
			
			LoadFromFile(true);

			XmlNode captureSourceNode = ConvertObjectToNode(config);

			captureSourceNodes[captureSourceOID] = captureSourceNode;

			SaveToFile();

			mtx.ReleaseMutex();
		}

		public void Delete(int captureSourceOID)
		{
			Logger.Verbose("RecorderHelperConfig: Delete(" + captureSourceOID.ToString() + ")");

			Mutex mtx = new Mutex(false, "RecorderHelper");
			if (!mtx.WaitOne(5000, false))
				throw new ApplicationException("Could not acquire RecorderHelper mutex");
			
			LoadFromFile(true);
			captureSourceNodes.Remove(captureSourceOID);
			SaveToFile();

			mtx.ReleaseMutex();
		}

		#endregion

		#region Private Node, Object Conversion Methods
		
		private object CopyNodeToObject(XmlNode node, object value)
		{
			Type valueType = value.GetType();

			FieldInfo[] fields = valueType.GetFields(BindingFlags.Instance | BindingFlags.NonPublic);
			
			object hashKey = null;
			foreach (FieldInfo field in fields) 
			{
				if ((field.Attributes & FieldAttributes.NotSerialized) != FieldAttributes.NotSerialized)
				{
					XmlArrayAttribute[] xmlArrays = (XmlArrayAttribute[])field.GetCustomAttributes(typeof(XmlArrayAttribute), false);
					XmlArrayItemAttribute[] xmlArrayItems = (XmlArrayItemAttribute[])field.GetCustomAttributes(typeof(XmlArrayItemAttribute), false);
					XmlAttributeAttribute[] xmlAttributes = (XmlAttributeAttribute[])field.GetCustomAttributes(typeof(XmlAttributeAttribute), false);
					XmlElementAttribute[] xmlElements = (XmlElementAttribute[])field.GetCustomAttributes(typeof(XmlElementAttribute), false);
					XmlHashKeyAttribute[] xmlHashKeys = (XmlHashKeyAttribute[])field.GetCustomAttributes(typeof(XmlHashKeyAttribute), false);

					object fieldValue = null;
					if ((xmlArrays.Length != 0) && (xmlArrayItems.Length != 0))
					{
						string arrayName;
						if ((xmlElements.Length != 0) && (xmlElements[0].ElementName.Length > 0))
							arrayName = xmlElements[0].ElementName;
						else
							arrayName = field.Name;
						XmlNode arrayElement = node.SelectSingleNode(arrayName);
						if (arrayElement != null)
						{
							Logger.Verbose("RecorderHelperConfig: LoadArray: " + arrayName);
							if (typeof(IDictionary).IsAssignableFrom(field.FieldType))
							{
								Hashtable newHashtable = new Hashtable();
								foreach (XmlNode subnode in arrayElement.ChildNodes)
								{
									Logger.Verbose("RecorderHelperConfig: LoadDictionaryItem: " + xmlArrayItems[0].Type.Name);
									object subvalue = xmlArrayItems[0].Type.GetConstructor(Type.EmptyTypes).Invoke(null);
									object subkey = CopyNodeToObject(subnode, subvalue);
									newHashtable.Add(subkey, subvalue);
								}
								fieldValue = newHashtable;
							}
							else
							{
								ArrayList newArrayList = new ArrayList();
								foreach (XmlNode subnode in arrayElement.ChildNodes)
								{
									Logger.Verbose("RecorderHelperConfig: LoadArrayItem: " + xmlArrayItems[0].Type.Name);
									object subvalue = xmlArrayItems[0].Type.GetConstructor(Type.EmptyTypes).Invoke(null);
									CopyNodeToObject(subnode, subvalue);
									newArrayList.Add(subvalue);
								}
								fieldValue = newArrayList;
							}
						}
						else
						{
							Logger.Verbose("RecorderHelperConfig: LoadArray: Missing!: " + arrayName);
						}
					}
					else if (xmlAttributes.Length != 0)
					{
						string attributeName;
						if (xmlAttributes[0].AttributeName.Length > 0)
							attributeName = xmlAttributes[0].AttributeName;
						else
							attributeName = field.Name;
						XmlAttribute fieldAttribute = node.Attributes[attributeName];
						if (fieldAttribute != null)
						{
							Logger.Verbose("RecorderHelperConfig: LoadAttribute: " + attributeName + "=" + fieldAttribute.Value);
							fieldValue = fieldAttribute.Value.ToString();
						}
						else
						{
							Logger.Verbose("RecorderHelperConfig: LoadAttribute: Missing!: " + attributeName);
						}
					}
					else
					{
						string elementName;
						if ((xmlElements.Length != 0) && (xmlElements[0].ElementName.Length > 0))
							elementName = xmlElements[0].ElementName;
						else
							elementName = field.Name;
						XmlNode fieldElement = node.SelectSingleNode(elementName);
						if (fieldElement != null)
						{
							Logger.Verbose("RecorderHelperConfig: LoadElement: " + elementName + "=" + fieldElement.InnerText);
							fieldValue = fieldElement.InnerText;
						}
						else
						{
							Logger.Verbose("RecorderHelperConfig: LoadElement: Missing!: " + elementName);
						}
					}

					if (fieldValue != null)
					{
						if (typeof(ICollection).IsAssignableFrom(field.FieldType))
						{
							field.SetValue(value, fieldValue);
						}
						else if (field.FieldType.IsEnum)
						{
							field.SetValue(value, Enum.Parse(field.FieldType, fieldValue.ToString()));
						}
						else if (typeof(Guid).IsAssignableFrom(field.FieldType))
						{
							field.SetValue(value, new Guid(fieldValue.ToString()));
						}
						else
						{
							field.SetValue(value, Convert.ChangeType(fieldValue, field.FieldType));
						}

						if (xmlHashKeys.Length > 0)
						{
							hashKey = field.GetValue(value);
						}
					}
				}
			}
			return hashKey;
		}

		private XmlNode ConvertObjectToNode(object value)
		{
			Type valueType = value.GetType();

			XmlTypeAttribute[] xmlTypes = (XmlTypeAttribute[])valueType.GetCustomAttributes(typeof(XmlTypeAttribute), false);

			XmlNode node;
			if ((xmlTypes.Length > 0) && (xmlTypes[0].TypeName != null))
				node = doc.CreateElement(xmlTypes[0].TypeName);
			else
				node = doc.CreateElement(valueType.Name);

			FieldInfo[] fields = valueType.GetFields(BindingFlags.Instance | BindingFlags.NonPublic);
			
			foreach (FieldInfo field in fields) 
			{
				object nodeValue = field.GetValue(value);
				if (((field.Attributes & FieldAttributes.NotSerialized) != FieldAttributes.NotSerialized) && (nodeValue != null))
				{
					XmlArrayAttribute[] xmlArrays = (XmlArrayAttribute[])field.GetCustomAttributes(typeof(XmlArrayAttribute), false);
					XmlAttributeAttribute[] xmlAttributes = (XmlAttributeAttribute[])field.GetCustomAttributes(typeof(XmlAttributeAttribute), false);
					XmlElementAttribute[] xmlElements = (XmlElementAttribute[])field.GetCustomAttributes(typeof(XmlElementAttribute), false);

					if (xmlArrays.Length != 0)
					{
						string arrayName;
						if ((xmlElements.Length != 0) && (xmlElements[0].ElementName.Length > 0))
							arrayName = xmlElements[0].ElementName;
						else
							arrayName = field.Name;
						Logger.Verbose("RecorderHelperConfig: SaveArray: " + arrayName);

						XmlElement arrayElement = doc.CreateElement(arrayName);

						if (typeof(IDictionary).IsAssignableFrom(field.FieldType))
						{
							foreach (DictionaryEntry subvalue in nodeValue as IEnumerable)
							{
								Logger.Verbose("RecorderHelperConfig: SaveDictionaryItem: " + subvalue.GetType().Name);
								arrayElement.AppendChild(ConvertObjectToNode(subvalue.Value));
							}
						}
						else
						{
							foreach (object subvalue in nodeValue as IEnumerable)
							{
								Logger.Verbose("RecorderHelperConfig: SaveArrayItem: " + subvalue.GetType().Name);
								arrayElement.AppendChild(ConvertObjectToNode(subvalue));
							}
						}

						node.AppendChild(arrayElement);
					}
					else if (xmlAttributes.Length != 0)
					{
						string attributeName;
						if (xmlAttributes[0].AttributeName.Length > 0)
							attributeName = xmlAttributes[0].AttributeName;
						else
							attributeName = field.Name;
						XmlAttribute fieldAttribute = doc.CreateAttribute(attributeName);
						fieldAttribute.Value = (string)nodeValue.ToString();
						node.Attributes.Append(fieldAttribute);
						Logger.Verbose("RecorderHelperConfig: SaveAttribute: " + fieldAttribute.Name + "=" + fieldAttribute.Value);
					}
					else
					{
						string elementName;
						if ((xmlElements.Length != 0) && (xmlElements[0].ElementName.Length > 0))
							elementName = xmlElements[0].ElementName;
						else
							elementName = field.Name;
						XmlElement fieldElement = doc.CreateElement(elementName); 
						fieldElement.InnerText = (string)nodeValue.ToString();
						node.AppendChild(fieldElement);
						Logger.Verbose("RecorderHelperConfig: SaveElement: " + fieldElement.Name + "=" + fieldElement.InnerText);
					}

				}
			}

			return node;
		}

		#endregion

		#region Private Load / Save File Methods

		private void LoadFromFile(bool reload)
		{
			if (!reload && (doc != null)) return;

			Logger.Verbose("RecorderHelperConfig: LoadFromFile()");

			doc = new XmlDocument();

			if (File.Exists(GetConfigFileName()))
				doc.Load(GetConfigFileName());

			globalSettingNodes = new HybridDictionary();

			foreach (XmlNode globalSettingsNode in doc.SelectNodes("settings/globalSettings/*"))
			{
				globalSettingNodes[globalSettingsNode.Name] = globalSettingsNode;
			}
           
			captureSourceNodes = new HybridDictionary();

			foreach (XmlNode captureSourceNode in doc.SelectNodes("settings/captureSources/captureSource"))
			{
				int captureSourceOID = Convert.ToInt32(captureSourceNode.Attributes.GetNamedItem("oid").Value);
                captureSourceNodes[captureSourceOID] = captureSourceNode;
			}
           
		}

		private void SaveToFile()
		{
			Logger.Verbose("RecorderHelperConfig: SaveToFile()");

			doc.RemoveAll();

			doc.AppendChild(doc.CreateXmlDeclaration("1.0", "UTF-8", null));

			XmlNode settings = doc.CreateElement("settings");
			doc.AppendChild(settings);

			XmlNode global = doc.CreateElement("globalSettings");
			settings.AppendChild(global);

			ArrayList globalSettingsKeys = new ArrayList(globalSettingNodes.Keys);
			globalSettingsKeys.Sort();
			foreach (string globalSettingKey in globalSettingsKeys)
			{
				global.AppendChild((XmlNode)globalSettingNodes[globalSettingKey]);
			}

			XmlNode sources = doc.CreateElement("captureSources");
			settings.AppendChild(sources);

			ArrayList captureSourcesKeys = new ArrayList(captureSourceNodes.Keys);
			captureSourcesKeys.Sort();
			foreach (int captureSourceOID in captureSourcesKeys)
			{
				sources.AppendChild((XmlNode)captureSourceNodes[captureSourceOID]);
			}

			doc.Save(GetConfigFileName());
		}

		private string GetConfigFileName()
		{
			if (configFileName == null) 
			{
				RegistryKey devnzKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\devnz\");
				string installDir = (string) devnzKey.GetValue("GBPVR InstallDir");
				devnzKey.Close();
				configFileName = Path.Combine(installDir, "config-recorders.xml");
			}
			return configFileName;
		}
		
		#endregion

		#region Singleton Implementation

		/// <summary> the global static configuration. </summary>
		[NonSerialized()]
		private static Config config = new Config();

		public static Config getConfig()
		{
			return config;
		}

		#endregion

	}

	#region Enhanced XML Attributes
	[AttributeUsage(AttributeTargets.Property | AttributeTargets.Field)]
	public class XmlHashKeyAttribute : Attribute
	{
	}
	#endregion

}
